/* pvfs-test.c
 *
 * This is derived from code given to me by Rajeev Thakur.  Dunno where
 * it originated.
 *
 * It's purpose is to produce aggregate bandwidth numbers for varying
 * block sizes, number of processors, an number of iterations.
 *
 * Compile time defines determine whether or not it will use mpi, while
 * all other options are selectable via command line.
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <error.h>
#include <string.h>
#include <sys/time.h>
#include <pvfs.h>
#include <pvfs_proto.h>
#ifdef __USE_MPI__
#include <mpi.h>
#endif


/* DEFAULT VALUES FOR OPTIONS */
int64_t opt_block     = 1048576*16;
int     opt_iter      = 1;
int     opt_stripe    = -1;
int     opt_correct   = 0;
int     opt_vfs       = 0;
int     amode         = O_RDWR | O_CREAT;
char    opt_file[256] = "/foo/test.out\0";
char    opt_pvfstab[256] = "notset\0";
int		opt_pvfstab_set = 0;

/* function prototypes */
int parse_args(int argc, char **argv);
double Wtime(void);

extern fdesc_p pfds[];
extern int errno;
extern int debug_on;

/* globals needed for getopt */
extern char *optarg;
extern int optind, opterr;

int main(int argc, char **argv)
{
	char *buf, *tmp, *buf2, *tmp2, *check;
	int fd, i, j, mynod=0, nprocs=1, err, my_correct = 1, correct, myerrno;
	double stim, etim;
	double write_tim = 0;
	double read_tim = 0;
	double read_bw, write_bw;
	double max_read_tim, max_write_tim;
	double min_read_tim, min_write_tim;
	double ave_read_tim, ave_write_tim;
	struct pvfs_stat pstat = {0,-1,16384,0,0};
	int64_t iter_jump = 0;

	/* startup MPI and determine the rank of this process */
#ifdef __USE_MPI__
	MPI_Init(&argc,&argv);
	MPI_Comm_size(MPI_COMM_WORLD, &nprocs);
	MPI_Comm_rank(MPI_COMM_WORLD, &mynod);
#endif

	/* parse the command line arguments */
	parse_args(argc, argv);
	pstat.ssize = opt_stripe;
	if(!opt_vfs){
		/* tell pvfs that we want to pass in metadata and have tcp
		 * connections setup for us at pvfs_open */
		amode = amode | O_META | O_CONN;
		if (mynod == 0) {
			printf("# Using native pvfs calls.\n");
		}
	}
	else{
		if (mynod == 0) {
			printf("# Using vfs calls.\n");
		}
	}
	
	/* kindof a weird hack- if the location of the pvfstab file was 
	 * specified on the command line, then spit out this location into
	 * the appropriate environment variable: */
	
	if	(opt_pvfstab_set){
		if((setenv("PVFSTAB_FILE", opt_pvfstab, 1)) < 0){
			perror("setenv");
			goto die_jar_jar_die;
		}
	}
	
	/* this is how much of the file data is covered on each iteration of
	 * the test.  used to help determine the seek offset on each
	 * iteration */
	iter_jump = nprocs * opt_block;
		
	/* setup a buffer of data to write */
	if (!(tmp = (char *) malloc(opt_block + 256))) {
		perror("malloc");
		goto die_jar_jar_die;
	}
	buf = tmp + 128 - (((long)tmp) % 128);  /* align buffer */

	if (opt_correct) {
		/* do the same buffer setup for verifiable data */
		if (!(tmp2 = (char *) malloc(opt_block + 256))) {
			perror("malloc2");
			goto die_jar_jar_die;
		}
		buf2 = tmp + 128 - (((long)tmp) % 128);
	}

	/* open the file for writing */
	if(opt_vfs){
		if ((fd = open(opt_file, amode, 0644)) < 0) {
			fprintf(stderr, "node %d, open error: %s\n", mynod, strerror(errno));
			goto die_jar_jar_die;
		}
	}
	else{
		if ((fd = pvfs_open(opt_file, amode, 0644, &pstat, NULL)) < 0) {
			fprintf(stderr, "node %d, open error: %s\n", mynod, strerror(errno));
			goto die_jar_jar_die;
		}

	}

	/* now repeat the seek and write operations the number of times
	 * specified on the command line */
	for (j=0; j < opt_iter; j++) {

		/* seek to an appropriate position depending on the iteration and
		 * rank of the current process */
		if(opt_vfs){
			lseek(fd, (j*iter_jump)+(mynod*opt_block), SEEK_SET);
		}
		else{
			pvfs_lseek(fd, (j*iter_jump)+(mynod*opt_block), SEEK_SET);
		}


		if (opt_correct) /* fill in buffer for iteration */ {
			for (i=mynod+j, check=buf; i<opt_block; i++,check++) *check=(char)i;
		}

		/* discover the starting time of the operation */
#ifdef __USE_MPI__
	   MPI_Barrier(MPI_COMM_WORLD);
	   stim = MPI_Wtime();
#else
		stim = Wtime();
#endif

		/* write out the data */
		if(opt_vfs){
			err = write(fd, buf, opt_block);
		}
		else{
			err = pvfs_write(fd, buf, opt_block);
		}
		myerrno = errno;

		/* discover the ending time of the operation */
#ifdef __USE_MPI__
	   etim = MPI_Wtime();
#else
		etim = Wtime();
#endif
	   write_tim += (etim - stim);

	   if (err < 0) fprintf(stderr, "node %d, write error, loc = %Ld: %s\n",
			mynod, mynod*opt_block, strerror(myerrno));
		
		/* we are done with this "write" iteration */
	}

	/* close the file */
	if(opt_vfs){
		if (close(fd) < 0) {
			fprintf(stderr, "node %d, close error after write\n", mynod);
		}
	}
	else{
		if (pvfs_close(fd) < 0) {
			fprintf(stderr, "node %d, close error after write\n", mynod);
		}
	}
	 
	/* wait for everyone to synchronize at this point */
#ifdef __USE_MPI__
	MPI_Barrier(MPI_COMM_WORLD);
#endif

	/* reopen the file to read the data back out */
	if(opt_vfs){
		if ((fd = open(opt_file, amode, 0644)) < 0) {
			fprintf(stderr, "node %d, open error: %s\n", mynod, strerror(errno));
			goto die_jar_jar_die;
		}
	}
	else{
		if ((fd = pvfs_open(opt_file, amode, 0644, &pstat, NULL)) < 0) {
			fprintf(stderr, "node %d, open error: %s\n", mynod, strerror(errno));
			goto die_jar_jar_die;
		}
	}

	/* we are going to repeat the read operation the number of iterations
	 * specified */
	for (j=0; j < opt_iter; j++) {

		/* seek to the appropriate spot give the current iteration and
		 * rank within the MPI processes */
		if(opt_vfs){
			lseek(fd, (j*iter_jump)+(mynod*opt_block), SEEK_SET);
		}
		else{
			pvfs_lseek(fd, (j*iter_jump)+(mynod*opt_block), SEEK_SET);
		}

		/* discover the start time */
#ifdef __USE_MPI__
	   MPI_Barrier(MPI_COMM_WORLD);
	   stim = MPI_Wtime();
#else
		stim = Wtime();
#endif

		/* read in the file data */
		if(opt_vfs){
			if (!opt_correct)
				err = read(fd, buf, opt_block);
			else
				err = read(fd, buf2, opt_block); /* read into second buffer */
		}
		else{
			if (!opt_correct)
				err = pvfs_read(fd, buf, opt_block);
			else
				err = pvfs_read(fd, buf2, opt_block); /* read into second buffer */
		}
		myerrno = errno;

		/* discover the end time */
#ifdef __USE_MPI__
	   etim = MPI_Wtime();
#else
		etim = Wtime();
#endif
	   read_tim += (etim - stim);

	   if (err < 0) fprintf(stderr, "node %d, read error, loc = %Ld: %s\n",
			mynod, mynod*opt_block, strerror(myerrno));

		/* if the user wanted to check correctness, compare the write
		 * buffer to the read buffer */
		if (opt_correct && memcmp(buf, buf2, opt_block)) {
			fprintf(stderr, "node %d, correctness test failed\n", mynod);
			my_correct = 0;
#ifdef __USE_MPI__
			MPI_Allreduce(&my_correct, &correct, 1, MPI_INT, MPI_MIN,
				MPI_COMM_WORLD);
#endif
		}

		/* we are done with this read iteration */
	}

	/* close the file */
	if(opt_vfs){
		pvfs_close(fd);
	}
	else{
		close(fd);
	}
	    
	/* compute the read and write times - not sure what the min_tim stuff
	 * is all about? -fil */
#ifdef __USE_MPI__
	MPI_Allreduce(&read_tim, &max_read_tim, 1, MPI_DOUBLE, MPI_MAX,
		MPI_COMM_WORLD);
	MPI_Allreduce(&read_tim, &min_read_tim, 1, MPI_DOUBLE, MPI_MIN,
		MPI_COMM_WORLD);
	MPI_Allreduce(&read_tim, &ave_read_tim, 1, MPI_DOUBLE, MPI_SUM,
		MPI_COMM_WORLD);
#else
	max_read_tim = read_tim;
	min_read_tim = read_tim;
	ave_read_tim = read_tim;
#endif
	/* calculate the average from the sum */
	ave_read_tim = ave_read_tim / nprocs; 

#ifdef __USE_MPI__
	MPI_Allreduce(&write_tim, &max_write_tim, 1, MPI_DOUBLE, MPI_MAX,
		MPI_COMM_WORLD);
	MPI_Allreduce(&write_tim, &min_write_tim, 1, MPI_DOUBLE, MPI_MIN,
		MPI_COMM_WORLD);
	MPI_Allreduce(&write_tim, &ave_write_tim, 1, MPI_DOUBLE, MPI_SUM,
		MPI_COMM_WORLD);
#else
	max_write_tim = write_tim;
	min_write_tim = write_tim;
	ave_write_tim = write_tim;
#endif
	/* calculate the average from the sum */
	ave_write_tim = ave_write_tim / nprocs; 
	
	/* print out the results on one node */
	if (mynod == 0) {
	   read_bw = ((int64_t)(opt_block*nprocs*opt_iter))/(max_read_tim*1000000.0);
	   write_bw = ((int64_t)(opt_block*nprocs*opt_iter))/(max_write_tim*1000000.0);
		
			printf("nr_procs = %d, nr_iter = %d, blk_sz = %ld\n", nprocs,
		opt_iter, (long)opt_block);
			
			printf("# total_size = %ld\n", (long)(opt_block*nprocs*opt_iter));
			
			printf("# Write:  min_time = %f, max_time = %f, ave_time = %f\n", 
				min_write_tim, max_write_tim, ave_write_tim);
			printf("# Read:  min_time = %f, max_time = %f, ave_time = %f\n", 
				min_read_tim, max_read_tim, ave_read_tim);
		
	   printf("Write bandwidth = %f Mbytes/sec\n", write_bw);
	   printf("Read bandwidth = %f Mbytes/sec\n", read_bw);
		
		if (opt_correct) {
			printf("Correctness test %s.\n", correct ? "passed" : "failed");
		}
	}


die_jar_jar_die:
	free(tmp);
	if (opt_correct) free(tmp2);
#ifdef __USE_MPI__
	MPI_Finalize();
#endif
	return(0);
}

int parse_args(int argc, char **argv)
{
	int c;
	
	while ((c = getopt(argc, argv, "s:b:i:f:p:cv")) != EOF) {
		switch (c) {
			case 's': /* stripe */
				opt_stripe = atoi(optarg);
				break;
			case 'b': /* block size */
				opt_block = atoi(optarg);
				break;
			case 'i': /* iterations */
				opt_iter = atoi(optarg);
				break;
			case 'f': /* filename */
				strncpy(opt_file, optarg, 255);
				break;
			case 'p': /* pvfstab file */
				strncpy(opt_pvfstab, optarg, 255);
				opt_pvfstab_set = 1;
				break;
			case 'c': /* correctness */
				opt_correct = 1;
				break;
			case 'v': /* use vfs interface */
				opt_vfs = 1;
				break;
			case '?': /* unknown */
			default:
				break;
		}
	}
	return(0);
}

/* Wtime() - returns current time in sec., in a double */
double Wtime()
{
	struct timeval t;
	
	gettimeofday(&t, NULL);
	return((double)t.tv_sec + (double)t.tv_usec / 1000000);
}





